@@ -17,6 +17,7 @@ else:
     from typing import Literal
 
 import aiodns
+import h11
 import websockets
 from websockets import WebSocketClientProtocol
 
@@ -215,7 +216,11 @@ class Client:
         for index, data in enumerate(first_packet):
             reader._buffer.insert(index, data)
 
-        if first_packet[0] == 4:  # Socks4
+        if not first_packet:
+            writer.close()
+            await writer.wait_closed()
+            return
+        elif first_packet[0] == 4:  # Socks4
             handler = getattr(self, "socks4")
         elif first_packet[0] == 5:  # Socks5
             handler = getattr(self, "socks5")
@@ -262,12 +267,57 @@ class Client:
         except OSError:
             logger.info(f"HTTP_{capwords(method)} ('{dsthost}', {dstport}) × (general)")
         else:
-            logger.info(f"HTTP_{capwords(method)} ('{dsthost}', {dstport}) √")
-            for index, data in enumerate(
-                (" ".join([method, urlpath, version]) + "\r\n").encode("ascii")
-            ):
+            proxy_or_direct = "PROXY" if self.is_proxyed(remote) else "DIRECT"
+            logger.info(
+                f"HTTP_{capwords(method)} ('{dsthost}', {dstport}) {proxy_or_direct} √"
+            )
+            for index, data in enumerate(firstline):
                 sock.r._buffer.insert(index, data)
-            await self.bridge(remote, sock)
+
+        server = h11.Connection(our_role=h11.SERVER)
+        client = h11.Connection(our_role=h11.CLIENT)
+        try:
+
+            async def server_task():
+                while True:
+                    server.receive_data(await sock.recv())
+                    event = server.next_event()
+                    logger.debug(f"HTTP Proxy Server: {event}")
+                    if event is h11.NEED_DATA:
+                        continue
+                    elif type(event) is h11.Request:
+                        event.target = (
+                            b"/"
+                            + event.target.split(b"://", maxsplit=1)[1].split(
+                                b"/", maxsplit=1
+                            )[1]
+                        )
+                        await remote.send(client.send(event))
+                    elif type(event) is h11.Data:
+                        await remote.send(client.send(event))
+                    elif type(event) in (h11.EndOfMessage, h11.ConnectionClosed):
+                        await remote.send(client.send(h11.EndOfMessage()))
+                        break
+                    else:
+                        break
+
+            async def client_task():
+                while True:
+                    client.receive_data(await remote.recv())
+                    event = client.next_event()
+                    logger.debug(f"HTTP Proxy Client: {event}")
+                    if event is h11.NEED_DATA:
+                        continue
+                    elif type(event) in (h11.Response, h11.Data):
+                        await sock.send(server.send(event))
+                    elif type(event) in (h11.EndOfMessage, h11.ConnectionClosed):
+                        await sock.send(server.send(h11.EndOfMessage()))
+                        break
+                    else:
+                        break
+
+            await onlyfirst(server_task(), client_task())
+        finally:
             await remote.close()
 
     async def http_connect(self, sock: TCPSocket) -> None:
@@ -303,7 +353,8 @@ class Client:
             logger.info(f"HTTP_Connect ('{dsthost}', {dstport}) × (general)")
         else:
             await reply(version, HTTPStatus.OK)
-            logger.info(f"HTTP_Connect ('{dsthost}', {dstport}) √")
+            proxy_or_direct = "PROXY" if self.is_proxyed(remote) else "DIRECT"
+            logger.info(f"HTTP_Connect ('{dsthost}', {dstport}) {proxy_or_direct} √")
             await self.bridge(remote, sock)
             await remote.close()
 
@@ -331,7 +382,8 @@ class Client:
         else:
             await sock.send(b"\x00\x90")
             await sock.send(data[2:8])
-            logger.info(f"Socks4_Connect ('{dsthost}', {dstport}) √")
+            proxy_or_direct = "PROXY" if self.is_proxyed(remote) else "DIRECT"
+            logger.info(f"Socks4_Connect ('{dsthost}', {dstport}) {proxy_or_direct} √")
             await self.bridge(remote, sock)
             await remote.close()
 
@@ -373,7 +425,8 @@ class Client:
         else:
             await sock.send(b"\x05\x00\x00")
             await sock.send(data[3:])
-            logger.info(f"Socks5_Connect ('{dsthost}', {dstport}) √")
+            proxy_or_direct = "PROXY" if self.is_proxyed(remote) else "DIRECT"
+            logger.info(f"Socks5_Connect ('{dsthost}', {dstport}) {proxy_or_direct} √")
             await self.bridge(remote, sock)
             await remote.close()
 
@@ -450,6 +503,9 @@ class Client:
             remote = await TCPSocket.create_connection(ip, port)
         return remote
 
+    def is_proxyed(self, connection: Socket) -> bool:
+        return isinstance(connection, WebSocket)
+
     @staticmethod
     async def bridge(s0: Socket, s1: Socket) -> None:
         async def _(sender: Socket, receiver: Socket):
